#PMS 20141127
#for a folder of Kerrmit difference images of DWs sweeping left to right  or vice versa(as an approx vertical line)
#finds the distance travelled in each image and saves this as a TDI format file with fields extracted from the image filenames
#also takes lnv, H-1/4 for creep regime analysis and works out statistical errors on averaged v
#version 5 extracts pulse lengths and pixel to micron conversion factors from filename if in form "sample_x20x_1spulse2300.png"
#and has metadata on how analysis is done and which image files were used

from Stoner.Folders import DataFolder
from Stoner import DataFile
from scipy import misc
from Stoner import Analysis
from Stoner import Plot
import numpy as np
from scipy import stats
import os

def linear(x,m,c):
    return m*x+c

def get_set_field(Sample, file_field):
    """Extracts the field setting voltage pulse from a file name of the form "ef03018_ip00000Oe_0600mV004s.png" (=2300mV), converts to a field 
    and returns the field value as a float.Accepts the filename as a string. 
    WHOOP coil. field in mT.
    """
    file_field=file_field.split(Sample + '_')[1] # chop of the start of the filename up to 'pulse' before voltage 
    fieldVpulse=float(file_field.split('mV')[0]) #chop off end of filename and change string to float
    set_field=fieldVpulse*0.02055
    return set_field

def get_pulse_length(file_name):
    """Extracts the field pulse length from a file name of the form "ef02911_*stuff*_0700mV008s.png" , 
    and returns the pulse length as a float.Accepts the filename as a string.
    """

    file_name=file_name.split('mV_')[1] # chop of the start of the filename up to 'x_' before pulse length
    
    if 's.' in file_name:
        file_name=file_name.split('s.')[0] #chop of the 2nd part of the filename
    else:
      file_name=file_name.split('s_')[0]
 
    try:
        pulse_length=float(file_name)
    except:
        decimal=file_name.split('p')
        pulse_length=float(decimal[0]+'.'+decimal[1])
    return pulse_length
	
def get_microns_per_pixel(file_name):
	"""Extracts the pixel to micron conversion factor from a file name of the form "sample_x20x_1spulse2300.png" , 
	and returns the value as a float.Accepts the filename as a string.
	"""
	file_name=file_name.split('_x')[1] # chop of the start of the filename up to 'x_' before pulse length 
	magnification=file_name.split('x_')[0] #chop of the 2nd part of the filename
	if magnification=='2p5':
		um_per_pixel=625/150
	elif magnification=='20':
		um_per_pixel = 0.667
	elif magnification=='50':
		um_per_pixel=50/150
	else:
		print ('Warning! Magnification from filename read as ', magnification,' . Set to 20x. If this is wrong, set manually and rerun')
		um_per_pixel=100/150 #take a guess at 20x mag if reading from filename doesn't work
	return um_per_pixel
 
def get_nucleation_centre (DWpic):
    ImageCentre = (int((DWpic.shape[0]-42)/2),int(DWpic.shape[1]/2))
    
    bins, edges = np.histogram(DWpic,bins=1000)
    
    brightlevel = edges[690:][np.argmax(bins[690:])]
    darklevel = edges[1:690][np.argmax(bins[1:690])]
    
    brightthreshold=darklevel+0.6*(brightlevel-darklevel)
    
    RightBound = ImageCentre[1]
    LeftBound = ImageCentre[1]
    UpperBound = ImageCentre[0]
    LowerBound = ImageCentre[0]    
    
    #Find upper boundry of nucleation site    
    for r in range(ImageCentre[0], -1, -1):
#        print('U'+str(r))
        #if DWtravel[r]<darkthreshold and DWtravel[r+2]<darkthreshold and DWtravel[r+4]<darkthreshold:#reversed region is dark  
        if DWpic[r, ImageCentre[1]]>brightthreshold and DWpic[r-2, ImageCentre[1]]>brightthreshold and DWpic[r-4, ImageCentre[1]] and DWpic[r-6, ImageCentre[1]]>brightthreshold: #reversed region is bright
            UpperBound=r
            break
        
    #Find lower boundry of nucleation site    
    for r in range(ImageCentre[0], len(DWpic[0])-42):
#        print('B'+str(r))
        #if DWtravel[r]<darkthreshold and DWtravel[r+2]<darkthreshold and DWtravel[r+4]<darkthreshold:#reversed region is dark  
        if DWpic[r, ImageCentre[1]]>brightthreshold and DWpic[r+2, ImageCentre[1]]>brightthreshold and DWpic[r+4, ImageCentre[1]] and DWpic[r+6, ImageCentre[1]]>brightthreshold: #reversed region is bright
            LowerBound=r
            break
        
    #Find leftmost boundry of nucleation site    
    for r in range(ImageCentre[1], -1, -1):
        #if DWtravel[r]<darkthreshold and DWtravel[r+2]<darkthreshold and DWtravel[r+4]<darkthreshold:#reversed region is dark  
        if DWpic[ImageCentre[0], r]>brightthreshold and DWpic[ImageCentre[0], r-2]>brightthreshold and DWpic[ImageCentre[0], r-4] and DWpic[ImageCentre[0], r-6]>brightthreshold: #reversed region is bright
            LeftBound=r
#            print('L' + str(LeftBound))
            break
        
    #Find rightmost boundry of nucleation site    
    for r in range(ImageCentre[1], len(DWpic[1])):
        #if DWtravel[r]<darkthreshold and DWtravel[r+2]<darkthreshold and DWtravel[r+4]<darkthreshold:#reversed region is dark  
        if DWpic[ImageCentre[0], r]>brightthreshold and DWpic[ImageCentre[0], r+2]>brightthreshold and DWpic[ImageCentre[0], r+4] and DWpic[ImageCentre[0], r+6]>brightthreshold: #reversed region is bright
            RightBound=r
#            print('R' + str(RightBound))
            break
        
    NucleationCentre = (UpperBound + (LowerBound-UpperBound)/2, LeftBound + (RightBound-LeftBound)/2)
    
    return NucleationCentre
	
def get_mean_DW_distance(y, DWpic, xstart, xend):
    """takes slices from left to right through an image and finds to distance DW has travelled
    """
    DWtravel=DWpic[y, xstart:xend] #slice out an array from the defect centre to the RHS of the picture #DWpic.shape gives shape in pixels as a tuple eg.(554, 672)
    ##this bit finds the start and end of the distance of DW motion from left to right (assuming the reversed region is the brighter part)
	
    bins, edges = np.histogram(DWpic,bins=1000)
    
    brightlevel = edges[690:][np.argmax(bins[690:])]
    darklevel = edges[1:690][np.argmax(bins[1:690])]
    
    brightthreshold=darklevel+0.6*(brightlevel-darklevel)
    darkthreshold=darklevel+0.4*(brightlevel-darklevel)

    DWstart=0
    DWend=0    
    
    try:
        for r in range(1, len(DWtravel)):
            #if DWtravel[r]<darkthreshold and DWtravel[r+2]<darkthreshold and DWtravel[r+4]<darkthreshold:#reversed region is dark 
            if DWtravel[r]>brightthreshold and DWtravel[r+2]>brightthreshold and DWtravel[r+4]>brightthreshold and DWtravel[r+6]>brightthreshold: #reversed region is bright
                DWstart=r
                break
        for r in range(DWstart, len(DWtravel)):
            #if DWtravel[r]>brightthreshold and DWtravel[r+2]>brightthreshold and DWtravel[r+4]>brightthreshold:#reversed region is dark 
            if DWtravel[r]<darkthreshold and DWtravel[r+2]<darkthreshold and DWtravel[r+4]<darkthreshold and DWtravel[r+6]<darkthreshold: #reversed region is bright
                DWend=r
                break
        d_pixel=DWend-DWstart
        
        if d_pixel==0:
            print('Something is wrong')
    except:
        d_pixel=0 #If there's a problem, just give 0. We can identify )s later and ignore them
        print ('oops')
#    print([d_pixel, DWstart, DWend])

    return d_pixel
 
def avsorterr(data,sort_col, ycols_errcols=None):
    """Sorts a data file so that data is in acsending order for chosen column
    (sortcol) keeping integrity of rows. Averages any duplicated points with same 
    x (i.e. sort_col) value and calculates standard errors on those points.
    
    To calculate standard errors, rather than just averging errors, 3rd argument 
    passed should be a list of pairs of y column indices and error column indices.
    So if there are two y columns (eg columns 1 and 5) with associated errors
    (eg columns 2 and 6) then pass ycols_errcols=[(1,2),(5,6)].
    """
    numberofcols=len(data.column_headers)
    sorted_data=DataFile()
    #get all of the x values
    x_list=[]
    for i in range(len(data.column(sort_col))):
        if data.column(sort_col)[i] not in x_list: 
            x_list.append(data.column(sort_col)[i])
    x_list.sort()
    for item in x_list:
        item_index=np.where(data.column(sort_col)==item)
        data_to_av=np.zeros(numberofcols)
        for index in item_index[0]: 
            data_to_av=np.vstack((data_to_av,np.array(data[int(index)])))
        data_to_av=np.delete(data_to_av,0,axis=0) #get rid of first row containing 0s
        data_row=np.mean(data_to_av, axis=0) #average each column
        if ycols_errcols: 
            stderrs=stats.sem(data_to_av, axis=0)
            for n in range(len(ycols_errcols)):
                data_row[(ycols_errcols[n][1])]=stderrs[(ycols_errcols[n][0])]
        sorted_data+=data_row
    return sorted_data


def ProcessData(Sample, path, output, microns_per_pixel, magnification_varies, pulsetime, pulse_varies, average_data):
    print(Sample)
    f=DataFolder(path, pattern='*.png') #these next few lines sort set up so we can iterate through the files
    ffile = f.files
    numberoffiles=len(ffile)
    print ('number of files=', numberoffiles)
    
    data=Analysis.AnalyseFile() #make a new file to add our analysed data to	
    
    #make some lists to add info that will be put in metadata at the end
    filenames_used=[]
    num_of_distances_used=[]
    pulse_lengths_used=[]
    um_per_pixel_used=[]
    
    for n in range(0, numberoffiles): #iterate through set of files to get distance domain wall travels and calculate parameters
        DWpic=misc.imread(ffile[n]) #read in image file
        NucleationCentre = get_nucleation_centre(DWpic)
        if output == LHS_output:
            xstart=0    #x pixel position to start iterating 
            xend=NucleationCentre[1]    #x pixel to stop at (max=672 i.e. RHS of image)
            ystart=NucleationCentre[0]-6  # where y starts
        elif output== RHS_output:
            xstart=NucleationCentre[1]  #x pixel position to start iterating 
            xend=len(DWpic[0])    #x pixel to stop at (max=672 i.e. RHS of image)
            ystart=NucleationCentre[0]-6  # where y starts
    
        y=ystart 
        total_d_pix=np.zeros(repeats) #array of zeros to fill with distances travelled
        for m in range(repeats): #gives repeats - number above
            y=y+pixelstep
            total_d_pix[m]=get_mean_DW_distance(y, DWpic, xstart, xend)#pass info to function to get distance and add to array            
            number_of_ds=np.count_nonzero(total_d_pix) #count hwo many speeds we've ended up with in the array (i.e. disregard 0s)
        if pulse_varies:
            pulsetime=get_pulse_length(ffile[n])
        if magnification_varies:
            microns_per_pixel=get_microns_per_pixel(ffile[n])
        v_micron=(np.sum(total_d_pix)/(pulsetime*number_of_ds))*microns_per_pixel #calculate velocity in microns/s
        v_micron_stat_error=(np.std(total_d_pix[total_d_pix !=0])/(pulsetime*np.sqrt(number_of_ds)))*microns_per_pixel
        v_micron_error=np.sqrt((microns_per_pixel/pulsetime)**2+v_micron_stat_error**2)	
        set_field=get_set_field(Sample, ffile[n]) #pass filename to function to get field from volatge pulse written in filename
        H1_4=1/(abs(set_field)**0.25) # field to the power of -1/4
        lnv=np.log(v_micron/1000000) #natural log of v. multiply by 1000 to put v into m/s
        lnv_error=v_micron_error/v_micron #multiply by 1000 to put v error into m/s
        data+=np.array([set_field,v_micron,v_micron_error,H1_4,lnv,lnv_error]) #add velocity data from this image to the output file
    	  #add info to metadata lists
        filenames_used=filenames_used+[ffile[n]]
        num_of_distances_used=num_of_distances_used+[number_of_ds]
        pulse_lengths_used=pulse_lengths_used+[pulsetime]
        um_per_pixel_used=um_per_pixel_used+[microns_per_pixel]
    
    if average_data:
        data=avsorterr(data,0, [(1,2),(4,5)])
        data.column_headers=data.column_headers
        data['averging']='sorted and averged using avsorterr'
       
    data.column_headers=["u0H (mT)","v (um/s)","+/-v","u0H^-1/4 (mT^-4)","ln(v)","+/-ln(v)"] #add column headers to the output file

    #add some metadata
    data['analysis code filename']=os.path.basename(__file__)
    data['repeats']=repeats
    data['pixelstep']=pixelstep
    data['filenames']=filenames_used
    data['no of distances used']=num_of_distances_used
    data['pulse length']=pulse_lengths_used
    data['um/pixel']=um_per_pixel_used
    print('hello')    
    analysis = Analysis.AnalyseFile(data)
    popt,pcov=analysis.curve_fit(linear,"u0H^-1/4 (mT^-4)", "ln(v)")
    print('hello1')	
#    err=np.sqrt(np.diag(pcov))
#    data['gradient']=popt[0]
#    data['gradient error']=err[0]
#    data['intercept']=popt[1]
#    data['intercept error']=err[1]
#    print (err)

#    data.save(output) 
    return data
#    print(data)

if __name__=="__main__":
    
    #switch automatic reading of pulse length and magnification off(=False) and on(=True) 
    pulse_varies = True # pulse length on/off
    pulsetime=1.0 #time field applied for in seconds. Set this if pulse_varies=0 (i.e. pulse length reading off)
    magnification_varies = False #micron per pixel on/off
    microns_per_pixel=1.3333  #conversion factor from pixels to microns. #2.5x2.5x =0.417
                                #For 2.5x lens = 625/150=4.167, 20x lens = 100/150=0.667, 50x lens = 50/150 =0.333
                                #for 10x lens = 200/150=1.33333...
    AverageData = True #Use avsorterr function to average and sort data
    
    repeats=30
    pixelstep=1
    
    LHS_Creep_Params=Analysis.AnalyseFile()
    RHS_Creep_Params=Analysis.AnalyseFile()
    
    
    for Sample in [('01',6),('03',7),('07',7.5), ('04',8), ('10',8.5), ('06',9), ('08',9.5), ('05',10), ('09',10.5)]:
        path='C:\\Users\\hazza\\Documents\\Summer2016\\Velocity measurements\\ef036_' + Sample[0] + '\\Velocity_Image\\diff\\' #change this to file path 
    
        LHS_output='C:\\Users\\hazza\\Documents\\Summer2016\\Velocity measurements\\ef036_' + Sample[0] + '\\Velocity_Image\\ef036_' + Sample[0] + '_LHS_20160627.txt' #change this to the file name you want the final data saved as
        RHS_output='C:\\Users\\hazza\\Documents\\Summer2016\\Velocity measurements\\ef036_' + Sample[0] + '\\Velocity_Image\\ef036_' + Sample[0] + '_RHS_20160627.txt' #change this to the file name you want the final data saved as
        

        LHS_results=ProcessData(Sample[0], path, LHS_output, microns_per_pixel, magnification_varies, pulsetime, pulse_varies, AverageData)
        RHS_results=ProcessData(Sample[0], path, RHS_output, microns_per_pixel, magnification_varies, pulsetime, pulse_varies, AverageData)
        print('step 2')
        
        LHS_Creep_Params+=np.array([Sample[0], Sample[1], LHS_results['gradient'], LHS_results['gradient error'], LHS_results['intercept'], LHS_results['intercept error']])
        RHS_Creep_Params+=np.array([Sample[0], Sample[1], RHS_results['gradient'], RHS_results['gradient error'], RHS_results['intercept'], RHS_results['intercept error']])
        
        LHS_results.save(LHS_output)
        RHS_results.save(RHS_output)
    
#        LHS_plot = Plot.PlotFile(LHS_results)
#        LHS_plot.plot_xy('u0H (mT)', 'v (um/s)', 'r.', yerr="+/-v", title='Left Hand Side Domain Wall Velocity Against Field')#, save_filename='Left_Hand_Side_Domain_Wall_Velocity_Against_Field')
#        LHS_plot2 = Plot.PlotFile(LHS_results)
#        LHS_plot2.plot_xy("u0H^-1/4 (mT^-4)", "ln(v)", 'r.', yerr="+/-ln(v)", title='Left Hand Side ln(v) Against u0H^-1/4(mT^-4)')#, save_filename='Left_Hand_Side_ln(v)_Against_H^-1/4(mT^-4)')
    
#        RHS_plot = Plot.PlotFile(RHS_results)
#        RHS_plot.plot_xy('u0H (mT)','v (um/s)','r.',yerr="+/-v",title='Right Hand Side Domain Wall Velocity Against Field')#, save_filename="Right_Hand_Side_Domain_Wall_Velocity_Against_Field")
#        RHS_plot2 = Plot.PlotFile(RHS_results)
#        RHS_plot2.plot_xy("u0H^-1/4 (mT^-4)","ln(v)",'r.',yerr="+/-ln(v)",title='Right Hand Side ln(v) Against H^-1/4(mT^-4)')#, save_filename='Right_Hand_Side_ln(v)_Against_H^-1/4(mT^-4)')
        
    LHS_Creep_Params.column_headers=["Sample", "Co Thickness", "Gradient", "Gradient Error", "Intercept", "Intercept Error"] #add column headers to the output file
    RHS_Creep_Params.column_headers=["Sample", "Co Thickness", "Gradient", "Gradient Error", "Intercept", "Intercept Error"] #add column headers to the output file

    LHS_Creep_Params.save('C:\\Users\\hazza\\Documents\\Summer2016\\Velocity measurements\\LHS_Creep_Params.txt')
    RHS_Creep_Params.save('C:\\Users\\hazza\\Documents\\Summer2016\\Velocity measurements\\RHS_Creep_Params.txt')